import logging

from . import SNSResource
from ..serializers import JSON


class Sink(SNSResource, JSON):
    """Uses the boto3 Topic object.

    Takes a topic object to put messages to it.

    Relies on create_topic() and topic.publish() as described in:
    http://boto3.readthedocs.org/en/latest/reference/services/sns.html#topic
    """

    #
    # TODO: use topic name for logging.
    #

    def __init__(self, topic=None, topic_name=None, sns=None, region_name=None, use_compression=True):
        """Prepare the sink with a configured SNS backend.

        Either the SNS topic object or its name must be passed (topic object is
        preferred).

        Keyword parameters:
        * topic: the topic object.
        * topic_name: alternatively, the topic name.
        * sns: optionaly, a SNS handler to get the topic by name.
        * region_name: optionally, the region name the SNS backend lives in.
        * use_compression: compress message using zlib before publishing (True by default)
        """
        self.use_compression = use_compression
        self.log = logging.getLogger('sns.sink')
        if topic is None:
            topic = self._get_topic(topic_name, sns, region_name)
        self.topic = topic
        self.log.debug('using SNS topic ARN="%s"' % self.topic.arn)

    def put(self, pkg):
        """Sends the raw package via the topic (to all transport protocols).

        Positional parameters:
        * the raw package dict.
        """
        msg = self.serialize(pkg)
        self.log.debug('sending to topic JSON encoded package: %s' % msg)
        if self.use_compression:
            msg = self.compress(msg)
        self.topic.publish(Message=msg)



"""
Top level utilities and documentation for the plumb implementation in AWS.
"""
import base64
import json
import zlib

import boto3
import botocore


class CompressionMixin:
    """Mixin to provide compression/decompression capablities."""

    DEFAULT_ENCODING = 'utf-8'

    def compress(self, message):
        return self._b64compress(message)

    def decompress(self, message):
        try:
            # There may be uncompressed messages in the queue
            json.loads(message)
            return message
        except json.decoder.JSONDecodeError:
            return self._b64decompress(message)

    def _b64compress(self, message):
        ''' Returns a base64-encoded string after compression.
            We always return a string, since SNS does not accept bytes. '''
        return base64.b64encode(zlib.compress(bytes(message, self.DEFAULT_ENCODING))).decode(
            self.DEFAULT_ENCODING)

    def _b64decompress(self, message):
        return zlib.decompress(base64.b64decode(message)).decode(self.DEFAULT_ENCODING)


class SQSResource(CompressionMixin):
    """Mixin to provide _get_queue() to fetch a SQS queue by name."""

    def _get_queue(self, queue_name, region_name=None, *, retry_delay=10):
        """Get the queue object by name and region. The function will retry to
        get the requested queue until it is returned.

        A Session for a queue and the SQS resource is always instantiated. The
        Session is never closed.

        looping until the requested queue is returned.

        Positional parameters:
        * the queue name.

        Keyword parameters:
        * retry_delay: the time delay (in seconds) to retry the request, in case it fails. The
        default value is 10.
        * the queue region: the default value is None; behaviour is given by boto3 module.
        Read the following documentation:
        http://boto3.readthedocs.org/en/latest/reference/core/session.html?highlight=session#module-boto3.session
        """
        session = boto3.session.Session(region_name=region_name)
        sqs = session.resource('sqs')
        try:
            queue = sqs.get_queue_by_name(QueueName=queue_name)
        except botocore.exceptions.ClientError:
            raise ValueError('could not connect to SQS queue "{}" in region "{}"'.format(
                queue_name, region_name))
        return queue


class SNSResource(CompressionMixin):
    """Mixin to provide _get_topic() to fetch a SNS topic by name."""

    def _get_topic(self, topic_name, sns=None, region_name=None):
        """Get the topic by name.

        A SNS resource handle can be passed, or one will be obtained with
        default configuration parameters.

        Positional parameters:
        * the topic name.
        * optionally, a SNS handler to get the topic by name.
        """
        if sns is None:
            if region_name is not None:
                session = boto3.session.Session(region_name=region_name)
                sns = session.resource('sns')
            else:
                sns = boto3.resource('sns')
        # Use property of SNS resource: create_topic() is idempotent.
        # http://boto3.readthedocs.org/en/latest/reference/services/sns.html#sns.Client.create_topic
        return sns.create_topic(Name=topic_name)



"""
Source and Sink implementation using Redis list.
"""

import logging
from ..serializers import JSON


class Source(JSON):
    """Use a Redis list as Source implementation.

    Uses a connected Redis client to pull packages from the backend list. It
    assumes UTF-8 character set for the Redis client.

    TODO: how do blocking/timeouts work?
    """

    def __init__(self, queue_name, *, redis_client=None):
        self.queue_name = queue_name
        self.client = redis_client
        self.log = logging.getLogger('redis.source.{}'.format(queue_name))
        self.log.debug('using Redis list "{}" as backend'.format(self.queue_name))

    def get(self, timeout=None):
        """Return a package, or None, after optional "timeout" seconds.

        If "timeout" is not provided (or is 0 or None), block forever.
        """
        # redis-py automatically handles convertion between None and 0 for
        # timeout, and between list of "list names" and a single name.
        result = self.client.brpop(self.queue_name, timeout=timeout)

        if result is not None:
            _, val = result
            result = self.deserialize(result[1].decode('utf-8'))

        return result


class Sink(JSON):
    """Use a Redis list as Sink backend.

    Uses a connected Redis client to push to the list. It assumes UTF-8
    character set for the Redis client.
    """

    def __init__(self, queue_name, *, redis_client=None):
        self.queue_name = queue_name
        self.client = redis_client
        self.log = logging.getLogger('redis.sink.{}'.format(queue_name))
        self.log.debug('using Redis list "{}" as backend'.format(self.queue_name))

    def put(self, pkg):
        """Put an object in the queue as a JSON string."""
        self.client.lpush(self.queue_name, self.serialize(pkg))
